__all__ = ["FastHealpix", "fast_healpix_mod"]
import os

try:
    from .fast_healpix import FastHealpix
    from .fast_healpix_c import lib as fast_healpix_mod

except ModuleNotFoundError:
    import distutils.sysconfig

    import cffi

    c_library_name = f"fast_healpix_c"

    c_function_defs = r"""
    void healpixls_to_radecdeg(long int *ihp, int n_hpxes, int Nside, double dx, double dy, double* ra, double* dec);
    void healpixl_to_radecdeg(long int hp, int Nside, double dx, double dy, double* ra, double* dec);
    void healpixl_grid_to_radecdeg(long int hp, int n_points, int Nside, double *dx, double *dy, double* ra, double* dec);
    long int radec_to_healpixl(double ra, double dec, int Nside);
    void radecs_to_healpixls(double *ra, double *dec, int n_points, long int *hpixes, int Nside);
    """

    # compile the C code and import the module
    ffi = cffi.FFI()
    ffi.set_source(
        c_library_name,
        open(f"{os.path.dirname(__file__)}/fast_healpix_c_code.c", "r").read(),
        extra_compile_args=["-O3", "-msse2", "-std=c99"],
    )
    ffi.cdef(c_function_defs)
    ffi.compile(
        verbose=True,
        target=f"{os.path.dirname(__file__)}/fast_healpix_c{distutils.sysconfig.get_config_var('EXT_SUFFIX')}",
        tmpdir=f"{os.path.dirname(__file__)}/",
    )

    from .fast_healpix import FastHealpix
    from .fast_healpix_c import lib as fast_healpix_mod



import builtins
import math
from collections import OrderedDict

import blosc
import numpy as np


class ArgusFITSParseError(Exception):
    pass


class ArgusFITSWriteError(Exception):
    pass


class ArgusFITSHeaderError(Exception):
    pass


class ArgusFitsHeader:
    def __init__(self):
        self.keyvals = OrderedDict()
        self.comments = OrderedDict()

    def __getitem__(self, key):
        return self.keyvals[key]

    def __setitem__(self, key, value):
        if type(value) is list and len(value) == 2:
            self.keyvals[key] = value[0]
            self.comments[key] = value[1]
        else:
            self.keyvals[key] = value

    def __iter__(self):
        return iter(self.keyvals)

    def keys(self):
        return self.keyvals.keys()

    def items(self):
        return self.keyvals.items()

    def values(self):
        return self.keyvals.values()

    def from_string(self, header_string):
        # header ends if END is the only thing on one line
        # returns True if we reached the end of the header, False otherwise
        got_header_end = False
        for line_start in range(0, len(header_string), 80):
            line = header_string[line_start : line_start + 80]

            if line.strip() == "END":
                got_header_end = True
                break
            else:
                keyword = line[0:8].strip()
                value = line[11:]
                s = value.split("/")
                value = s[0]
                if len(s) > 1:
                    comment = s[1].strip()
                else:
                    comment = None

                # see if we can convert to a bool or number
                if value.replace("'", "").strip() == "T":
                    value = True
                elif value.replace("'", "").strip() == "F":
                    value = False
                else:
                    try:
                        value = int(value)
                    except ValueError:
                        try:
                            value = float(value)
                        except ValueError:
                            value = value.replace('"', "").replace("'", "").strip()
                self.keyvals[keyword] = value
                if comment is not None:
                    self.comments[keyword] = value
        return got_header_end

    def update(self, other_header):
        self.keyvals.update(other_header.keyvals)
        self.comments.update(other_header.comments)


class ArgusFitsHDU:
    # only supports image HDUs
    def __init__(self):
        self.data = None
        self.header = ArgusFitsHeader()


class ArgusHDUList:
    def __init__(self, hdus=None):
        self.hdu_list = []

        if type(hdus) is list:
            self.hdu_list = hdus
        if type(hdus) is ArgusFitsHDU:
            self.hdu_list = [hdus]

    def append(self, hdu):
        self.hdu_list.append(hdu)

    def __len__(self):
        return len(self.hdu_list)

    def __repr__(self):
        return self.hdu_list.__repr__()

    def __getitem__(self, n):
        return self.hdu_list[n]

    def __iter__(self):
        self.n = 0
        return self

    def __next__(self):
        self.n += 1
        if self.n < len(self.hdu_list):
            return self.hdu_list[self.n]
        else:
            raise StopIteration

    def writeto(self, filename, overwrite=True):  # noqa: C901
        if not overwrite:
            raise ArgusFITSWriteError("Protecting file write not supported")

        if "XTENSION" in self.hdu_list[0].header:
            raise ArgusFITSWriteError("HDUList does not have a PrimaryHDU")

        with builtins.open(filename, "wb") as outfile:
            for hdu_n, hdu in enumerate(self.hdu_list):
                # write the header first
                if hdu.data is not None:
                    out_array = np.ascontiguousarray(hdu.data.copy())

                    # automatically put the required imaging keywords into the header
                    # and do some other format-dependent setup
                    if hdu.data.dtype == np.int16:
                        hdu.header["BITPIX"] = 16
                        bytes_per_pix = 2
                        out_array = out_array.byteswap()
                    elif hdu.data.dtype == np.uint16:
                        hdu.header["BITPIX"] = 16
                        hdu.header["BZERO"] = 32768
                        hdu.header["BSCALE"] = 1
                        bytes_per_pix = 2
                        out_array = (out_array.astype(np.int32) - 32768.0).astype(
                            np.int16
                        )
                        out_array = out_array.byteswap()
                    elif hdu.data.dtype == np.int32:
                        hdu.header["BITPIX"] = 32
                        bytes_per_pix = 4
                        out_array = out_array.byteswap()
                    elif hdu.data.dtype == np.float16:
                        hdu.header[
                            "SIMPLE"
                        ] = False  # tell readers we're doing something naughty
                        bytes_per_pix = 2
                        hdu.header["BITPIX"] = -16
                    elif hdu.data.dtype == np.float32:
                        bytes_per_pix = 4
                        hdu.header["BITPIX"] = -32
                        out_array = out_array.byteswap()
                    elif hdu.data.dtype == np.float64:
                        bytes_per_pix = 8
                        hdu.header["BITPIX"] = -64
                        out_array = out_array.byteswap()
                    else:
                        raise ArgusFITSWriteError(
                            "Unsupported numpy data type (%s) writing extension %i"
                            % (repr(hdu.data.dtype), hdu_n)
                        )

                    if len(hdu.data.shape) != 2:
                        raise ArgusFITSWriteError(
                            "Only 2D arrays are supported, extension %i has %i dimensions"
                            % (hdu_n, len(hdu.data.shape))
                        )

                    hdu.header["NAXIS"] = 2
                    hdu.header["NAXIS1"] = hdu.data.shape[1]
                    hdu.header["NAXIS2"] = hdu.data.shape[0]
                    hdu.header["PCOUNT"] = 0
                    hdu.header["GCOUNT"] = 1
                else:
                    hdu.header["SIMPLE"] = True
                    hdu.header["BITPIX"] = 8
                    hdu.header["NAXIS"] = 0
                    hdu.header["EXTEND"] = True

                header_string = ""

                # the header is an OrderedDict, so we'll rely on the keywords having been given
                # in the correct FITS order
                for key, value in hdu.header.items():
                    if type(value) is str:
                        val = value
                    elif type(value) is bool:
                        if value:
                            val = "T"
                        else:
                            val = "F"
                    else:
                        val = repr(value)

                    try:
                        comment = hdu.header.comments[key]
                    except KeyError:
                        comment = None

                    if key == "SIMPLE":
                        header_string += "%-80s" % (
                            "%-8s=%21s / conforms to FITS standard" % (key, val)
                        )
                    elif key == "XTENSION":
                        header_string += "%-80s" % ("XTENSION= 'IMAGE   '")

                    else:
                        # get past astropy checks
                        if comment is not None:
                            full_line = "%-8s=%21s / %s" % (key, val, comment)
                        else:
                            full_line = "%-8s=%21s" % (key, val)

                        if len(full_line) > 80:
                            full_line = full_line[:80]

                        header_string += "%-80s" % full_line

                header_string += "%-80s" % "END"

                best_header_size = ((len(header_string) // 2880) + 1) * 2880
                header_format_spec = "%%-%is" % best_header_size
                header_string = header_format_spec % header_string
                header = header_string.encode("utf8")

                outfile.write(header)

                if hdu.data is not None:
                    out_array.tofile(outfile)
                    len_written = hdu.data.shape[0] * hdu.data.shape[1] * bytes_per_pix

                    if len_written % 2880 != 0:
                        # write extra to cope with the FITS file format's chunking
                        extra_length = 2880 - (len_written % 2880)
                        np.zeros(
                            extra_length // bytes_per_pix, dtype=hdu.data.dtype
                        ).tofile(outfile)


def read_blosc_from(file, index, size, dtype=np.float16, compressed=True):
    with builtins.open(file, "rb") as f:
        f.seek(index)
        bytes_read = f.read(size)
        if compressed:
            bytes_read = blosc.decompress(bytes_read)
        data = np.frombuffer(bytes_read, dtype=dtype)
        res = int(math.sqrt(data.size))
        data = data.reshape((res, res))
    return data


def load_hdu_from_file_handle(  # noqa: C901
    file_handle, verbose=False, get_byte_positions=False
):
    # loads from open file until end of HDU,
    # then returns the handle for later seeking
    # requires the file to have been opened in "rb" mode
    got_header_end = False
    hdu = ArgusFitsHDU()
    if verbose:
        print("Loading HDU")

    while not got_header_end:
        header_string = file_handle.read(2880)

        if len(header_string) == 0:
            if get_byte_positions:
                return None, None, None, None
            return None, None

        if len(header_string) < 80:
            raise ArgusFITSParseError("Reached end of file before finishing header")

        header_string = str(header_string, "utf-8")
        if hdu.header.from_string(header_string) is True:
            got_header_end = True

    data_start = file_handle.tell()
    if verbose:
        print("Loaded header:", hdu.header.items())

    # now in image section, if there is one
    if hdu.header["NAXIS"] == 2:
        if "TTYPE1" in hdu.header:
            if hdu.header["TTYPE1"] == "COMPRESSED_DATA":
                raise ArgusFITSParseError(
                    "Image HDUs are compressed, use Astropy instead."
                )
        dtype = None
        byte_swap = True
        if "BLOSC" in hdu.header:
            blosc_data = hdu.header["BLOSC"]
        else:
            blosc_data = False

        if hdu.header["BITPIX"] == 16:
            dtype = np.int16
        if hdu.header["BITPIX"] == 32:
            dtype = np.int32
        if hdu.header["BITPIX"] == -16:
            dtype = np.float16
            byte_swap = False  # since half-float is off the FITS standard anyway...
        if hdu.header["BITPIX"] == -32:
            dtype = np.float32
        if hdu.header["BITPIX"] == -64:
            dtype = np.float64

        if dtype is None:
            raise ArgusFITSParseError(
                f"Unknown datatype {hdu.header['BITPIX']} NAxis: {hdu.header['NAXIS']}"
            )

        n_pixels = hdu.header["NAXIS1"] * hdu.header["NAXIS2"]

        try:
            if blosc_data:
                blosc_bytes = file_handle.read(hdu.header["BLOSCBY"])
                bytes_read = len(blosc_bytes)
                byte_string = blosc.decompress(blosc_bytes)
                hdu.data = np.frombuffer(byte_string, dtype=dtype)
            else:
                hdu.data = np.fromfile(file_handle, dtype=dtype, count=n_pixels)
                bytes_read = (
                    n_pixels * dtype(1).nbytes
                )  # the 1 instantiates an object, required to get attributes
        except ValueError:
            raise ArgusFITSParseError("Error reading image data")

        if verbose:
            print("Loaded", n_pixels, "with a data type of", repr(dtype))
        if byte_swap:
            hdu.data.byteswap(inplace=True)

        if (hdu.header["BITPIX"] > 0) and ("BZERO" in hdu.header):
            if hdu.header["BZERO"] != 0:
                hdu.data += hdu.header["BZERO"]

            if hdu.header["BSCALE"] != 1:
                hdu.data *= hdu.header["BSCALE"]

        hdu.data = hdu.data.reshape(hdu.header["NAXIS2"], hdu.header["NAXIS1"])

        # and read out the alignment blanks so we're in the right place for the next HDU

        if bytes_read % 2880 != 0:
            extra_bytes = 2880 - (bytes_read % 2880)
            file_handle.read(extra_bytes)

    else:
        data_start = 0
        bytes_read = 2880

    if get_byte_positions:
        return file_handle, hdu, data_start, bytes_read
    else:
        return file_handle, hdu


def open(filename, verbose=False, get_byte_positions=False):
    hdulist = ArgusHDUList()

    # need to do it this way to match astropy's module.open() syntax
    with builtins.open(filename, "rb") as file_handle:
        end_of_file = False
        while not end_of_file:
            if get_byte_positions:
                file_handle, hdu, byte_start, nbytes = load_hdu_from_file_handle(
                    file_handle, verbose, get_byte_positions=get_byte_positions
                )
            else:
                file_handle, hdu = load_hdu_from_file_handle(file_handle, verbose)

            if file_handle is not None:
                if get_byte_positions:
                    hdu.header["_BYTE_START"] = byte_start
                    hdu.header["_NBYTES"] = nbytes
                hdulist.append(hdu)
            else:
                end_of_file = True
    return hdulist


if __name__ == "__main__":
    # speed test (verification tests are in test directory)
    import time

    import astropy.io.fits as fits

    argus_hdulist = open("20170308.f32.fits")

    tm = time.perf_counter()
    for n in range(10):
        argus_hdulist = open("20170308.fits")
        for d in argus_hdulist[1:]:
            d.data = d.data.astype(np.float32)
        argus_hdulist.writeto("argus_test.fits", overwrite=True)
    print("Argus: %fs" % (time.perf_counter() - tm))

    tm = time.perf_counter()
    for n in range(10):
        hdulist = fits.open("20170308.fits")
        for d in hdulist[1:]:
            d.data = d.data.astype(np.float32)
        argus_hdulist.writeto("argus_test.fits", overwrite=True)
    print("Argus: %fs" % (time.perf_counter() - tm))

    hdulist = fits.HDUList(fits.PrimaryHDU())
    for dtype in [np.float32, np.float64, np.int16, np.int32, np.uint16]:
        test_array = np.ones((10, 10), dtype=dtype) * 1000
        hdulist.append(fits.ImageHDU(test_array))
        hdulist[-1].header["DTYPE"] = dtype.__name__
    hdulist.writeto("test_dtypes.fits", overwrite=True)

    argus_hdulist = open("test_dtypes.fits")
    print("Argus loaded values (should all be 1000 or 1000.0)")
    for hdu in argus_hdulist:
        print("\t", hdu.header["DTYPE"], hdu.data[0, 0])
    argus_hdulist.writeto("test_dtypes_argus.fits")

    # and load it back in
    print("Argus written values as loaded by astropy")
    argus_hdulist = fits.open("test_dtypes.fits")  # this does complete FITS parsing, so
    # tests both header & image parsing
    for hdu in argus_hdulist[1:]:
        print("\t", hdu.header["DTYPE"], hdu.data[0, 0])



import datetime
import os
import socket
import uuid
import warnings
from typing import Any, Dict, Optional, Type, TypeVar, Union

import astropy.io.fits as fits
import astropy.units as u
import astropy.wcs as awcs
import astropy_healpix as ahpx
import numpy as np
from geojson_pydantic import Polygon
from pydantic import BaseModel, Field, ValidationError

warnings.simplefilter("ignore", category=awcs.FITSFixedWarning)


# Q = qlsc.QLSC(depth=30)
H = ahpx.HEALPix(nside=2**24, order="nested")
HOSTNAME = socket.gethostname()


EVRImageType = TypeVar("EVRImageType", bound="EVRImage")
EVRImageUpdateType = TypeVar("EVRImageUpdateType", bound="EVRImageUpdate")


def _wcs_to_footprint(header: fits.Header) -> Dict[str, Any]:
    """
     Converts the WCS header information to a GeoJSON polygon footprint.

    Args:
        header: FITS header containing WCS information.

    Returns:
        Dictionary representing the GeoJSON polygon footprint.

    """
    w = awcs.WCS(header)
    im_center = w.all_pix2world(3288, 2192, 0)
    header["CRVAL1"] = float(im_center[0])
    header["CRVAL2"] = float(im_center[1])

    x = np.arange(0, 6575, 274)
    y = np.arange(0, 4384, 274)

    bottom_edge = list(zip([0] * len(x), x)) + [(0, 6575)]
    right_edge = list(zip(y, [6575] * len(y)))[1:] + [(4383, 6575)]
    top_edge = list(zip([4383] * len(x), x)) + [(4383, 6575)]
    top_edge = list(reversed(top_edge))[1:]
    left_edge = list(zip(y, [0] * len(y))) + [(4383, 0)]
    left_edge = list(reversed(left_edge))[1:]

    poly = np.array(bottom_edge + right_edge + top_edge + left_edge)
    poly = np.array([poly.T[1], poly.T[0]]).T

    radecs = w.all_pix2world(poly, 0)
    # Re-orient RA from [-180,180]
    radecs[:, 0] -= 180
    radecs = [list(r) for r in radecs]

    footprint = {"type": "Polygon", "coordinates": [radecs]}
    return footprint


def fitspath_to_constructor(fitspath: Union[str, fits.HDUList]) -> Dict[str, Any]:
    """
    Convert a FITS header into a Pydantic constructor dictionary.

    Args:
        fitspath (str or astropy.io.fits.HDUList): The path to the FITS file or
        and HDUList object.

    Returns:
        Dict[str, Any]: The Pydantic constructor dictionary representing the FITS header.

    Raises:
        KeyError: If any required FITS header fields are missing.
    """
    if type(fitspath) is str:
        hdulist = fits.open(fitspath)

    if len(hdulist) == 1:
        hdu_num = 0
    else:
        hdu_num = 1
    header = hdulist[hdu_num].header.copy()
    hdulist.close()

    if "FIELDID" not in header:
        header["FIELDID"] = 400

    constructor_dict = {}
    constructor_dict["camera"] = header["CCDDETID"]
    constructor_dict["filter_name"] = header["FILTER"]
    constructor_dict["image_type"] = header["IMGTYPE"].strip()
    constructor_dict["obstime"] = header["DATE-OBS"] + "T" + header["TIME-OBS"]
    if "GPSTIME" in header and "Z" in header["GPSTIME"]:
        constructor_dict["gpstime"] = header["GPSTIME"]
    constructor_dict["ccd_set_temp"] = header["SETTEMP"]
    constructor_dict["ccd_temp"] = header["CCDTEMP"]
    constructor_dict["exp_time"] = header["EXPTIME"]
    constructor_dict["site_name"] = header["TELOBS"]
    constructor_dict["server_name"] = HOSTNAME
    constructor_dict["basename"] = header["ORIGNAME"].split(".")[0]
    constructor_dict["fieldid"] = header["FIELDID"]
    constructor_dict["ratchnum"] = header["RATCHNUM"]
    constructor_dict["mount_ha"] = header["MOUNTHA"]
    constructor_dict["sha1"] = header["CHECKSUM"]
    constructor_dict["ccd_ext_temp"] = header["CCDETEMP"]
    constructor_dict["wind_dir"] = header["WINDDIR"]
    constructor_dict["wind_speed"] = header["WINDSPED"]
    constructor_dict["rel_humidity"] = header["OUTRELHU"]
    constructor_dict["dew_point"] = header["OUTDEWPT"]
    constructor_dict["air_pressure"] = header["OUTPRESS"]
    constructor_dict["mushroom_temp"] = header["INR1TEMP"]
    constructor_dict["inc_x"] = header["INR1INCX"]
    constructor_dict["inc_y"] = header["INR1INCY"]
    constructor_dict["inc_z"] = header["INR1INCZ"]

    if "CRVAL1" in header:
        constructor_dict["hid"] = H.lonlat_to_healpix(
            header["CRVAL1"] * u.deg, header["CRVAL2"] * u.deg
        )

        constructor_dict["footprint"] = Polygon(**_wcs_to_footprint(header))
        constructor_dict["wcspath"] = os.path.abspath(fitspath)
    else:
        constructor_dict["rawpath"] = os.path.abspath(fitspath)
    return constructor_dict


class EVRImage(BaseModel):
    """Class representing an EVR image."""

    id: str = Field(default_factory=uuid.uuid4, alias="_id")
    camera: Optional[str] = Field(...)
    filter_name: str = Field(...)
    obstime: datetime.datetime = Field(...)
    gpstime: Optional[datetime.datetime] = None
    ccd_set_temp: float = Field(...)
    ccd_temp: float = Field(...)
    exp_time: float = Field(...)

    site_name: str = Field(...)
    server_name: str = Field(...)
    image_type: str = Field(...)
    rawpath: Optional[str] = None
    wcspath: Optional[str] = None
    basename: str = Field(...)
    fieldid: float = Field(...)
    ratchnum: str = Field(...)
    mount_ha: Optional[float] = Field(...)
    sha1: str = Field(...)
    ccd_ext_temp: float = Field(...)
    wind_dir: float = Field(...)
    wind_speed: float = Field(...)
    rel_humidity: float = Field(...)
    dew_point: float = Field(...)
    air_pressure: float = Field(...)
    mushroom_temp: float = Field(...)
    inc_x: Optional[float] = Field(...)
    inc_y: Optional[float] = Field(...)
    inc_z: Optional[float] = Field(...)

    footprint: Optional[Polygon] = Field(...)
    hid: Optional[int] = Field(...)

    class Config:
        """Pydantic configuration."""

        allow_population_by_field_name = True

    @classmethod
    def from_fits(cls: Type[EVRImageType], fitspath: str) -> EVRImageType:
        """Create an instance of EVRImage from a FITS file.

        Args:
            fitspath: Path to the FITS file.

        Returns:
            An instance of EVRImage.

        """

        constructor_dict = fitspath_to_constructor(fitspath)

        # This is duplicated boiler plate between the Image/ImageUpdate classes
        try:
            inst = cls(**constructor_dict)
        except ValidationError as e:
            for error in e.errors():
                field = error["loc"][0]
                constructor_dict[field] = None
            inst = cls(**constructor_dict)

        return inst


class EVRImageUpdate(BaseModel):
    """Class representing an update to an EVR image."""

    camera: Optional[str]
    filter_name: Optional[str]
    obstime: Optional[datetime.datetime]
    gpstime: Optional[datetime.datetime]
    ccd_set_temp: Optional[float]
    ccd_temp: Optional[float]
    exp_time: Optional[float]
    site_name: Optional[str]
    server_name: Optional[str]
    image_type: Optional[str]
    rawpath: Optional[str]
    wcspath: Optional[str]
    basename: Optional[str]
    fieldid: Optional[float]
    ratchnum: Optional[str]
    sha1: Optional[str]
    ccd_ext_temp: Optional[float]
    wind_dir: Optional[float]
    wind_speed: Optional[float]
    rel_humidity: Optional[float]
    dew_point: Optional[float]
    air_pressure: Optional[float]
    mushroom_temp: Optional[float]
    inc_x: Optional[float]
    inc_y: Optional[float]
    inc_z: Optional[float]
    footprint: Optional[Polygon]
    hid: Optional[int]

    class Config:
        """Pydantic configuration."""

        allow_population_by_field_name = True

    @classmethod
    def from_fits(cls: Type[EVRImageUpdateType], fitspath: str) -> EVRImageUpdateType:
        """Create an instance of EVRImageUpdate from a FITS file.

        Args:
            fitspath: Path to the FITS file.

        Returns:
            An instance of EVRImageUpdate.

        """
        constructor_dict = fitspath_to_constructor(fitspath)

        # This is duplicated boiler plate between the Image/ImageUpdate classes
        try:
            inst = cls(**constructor_dict)
        except ValidationError as e:
            for error in e.errors():
                field = error["loc"][0]
                constructor_dict[field] = None
            inst = cls(**constructor_dict)

        return inst



__all__ = [
    "EVRImage",
    "EVRImageUpdate",
    "EVRImageType",
    "EVRImageUpdateType",
    "fitspath_to_constructor",
    "EFTEAlert",
    "EFTECandidate",
    "XMatchItem",
]

from .efte_alert import EFTEAlert, EFTECandidate, XMatchItem
from .evr_image import EVRImage, EVRImageUpdate, fitspath_to_constructor



__all__ = ["S3AuthException", "S3Share"]
"""S3 Share module for handling Wasabi S3 interactions."""
import os
import tarfile
from typing import NoReturn

import boto3

from . import config


class S3AuthException(Exception):
    """Exception raised for S3 authentication errors."""

    pass


class S3Share:
    def __init__(self) -> None:
        """Initialize the S3Share class.

        Raises:
            S3AuthException: If Wasabi credentials are missing in ricoenv.
        """
        if config.WASABI_KEY_ID is None or config.WASABI_SECRET_KEY is None:
            raise S3AuthException("No Wasabi credentials in ricoenv")

        self.s3 = boto3.resource(
            "s3",
            aws_access_key_id=config.WASABI_KEY_ID,
            aws_secret_access_key=config.WASABI_SECRET_KEY,
            endpoint_url=config.WASABI_ENDPOINT,
        )

        self.vetnet_bucket = self.s3.Bucket("efte.vetnet")
        self.catalog_bucket = self.s3.Bucket("hera.catalogs")
        self.stamps_bucket = "efte.stamps"

    def upload_stamp(self, id: str) -> str:
        """
        Upload a stamp for to S3.

        Args:
            id (str): S3 key for the stamp (candidate ID)

        Returns:
            str: Pre-authenticated URL for the stamp file.
        """
        url = self.s3.generate_presigned_url(
            ClientMethod="get_object",
            Params={"Bucket": self.stamps_bucket, "Key": "invoice.pdf"},
        )
        return url

    def download_vetnet(self) -> NoReturn:
        """Download Vetnet data from S3.

        Downloads the hypersky_v7_v0.tar.gz file from the S3 bucket,
        extracts its contents, and saves it in the cache directory.

        Raises:
            botocore.exceptions.ClientError: If the file download fails.
        """
        cached_path = os.path.join(
            os.path.expanduser("~"), ".rico_cache", "hypersky_v7_v0.tar.gz"
        )

        self.vetnet_bucket.download_file(
            "hypersky_v7_v0.tar.gz",
            cached_path,
        )
        outpath = os.path.dirname(cached_path)
        file = tarfile.open(cached_path)
        file.extractall(outpath)
        file.close()

    def download_atlas(self) -> NoReturn:
        """Download ATLAS-RefCat 2 data from S3.

        Downloads the atlas_feathercat.tar.gz file from the S3 bucket,
        extracts its contents, and saves it in the cache directory.

        Raises:
            botocore.exceptions.ClientError: If the file download fails.
        """
        cached_path = os.path.join(config.RICO_CACHE_DIR, "atlas_feathercat.tar.gz")

        self.catalog_bucket.download_file(
            "atlas_feathercat.tar.gz",
            cached_path,
        )
        outpath = os.path.dirname(cached_path)
        file = tarfile.open(cached_path)
        file.extractall(outpath)
        file.close()



import glob
import multiprocessing as mp
import operator
import os
import warnings
from functools import reduce

import astropy.coordinates as acrd
import astropy_healpix as ahpx
import click
import numpy as np
import orjson
import pyarrow as pa
import pyarrow.parquet as pq

from . import afits as arf
from . import config, fast_healpix


class SegmentArchiveException(Exception):
    """Exception raised SkyMap error.

    Attributes:
        message -- explanation of the error
    """

    def __init__(self, message="Segment archive is corrupted."):
        self.message = message
        super().__init__(self.message)


class SegmentArchive:
    def __init__(self, root, load_index=True):
        self.root = root

        # Note: This should be pulled from a manifest in the root directory
        # of the data archive, which is not currently implemented.
        self.manifest = None
        self.nside_base = 32

        if self.nside_base == 32:  # Evryscope
            self.subdivs = [self.nside_base // 16, self.nside_base // 4]
        elif self.nside_base == 256:  # Argus
            self.subdivs = [
                self.nside_base // 64,
                self.nside_base // 16,
                self.nside_base // 4,
            ]
        else:
            raise SegmentArchiveException(
                f"Base NSide of SkyMap must be 32 (EVR) or 256 (Argus), got {self.nside_base}"
            )

        # NB: The following astropy lookups are used for pixel metadata
        # Actual geometry is done with the fast_healpix subpackage.
        self.hpix = ahpx.HEALPix(
            nside=self.nside_base,
            order="nested",
            frame=acrd.ICRS(),
        )

        self.subdiv_hpx = []
        for nside in self.subdivs:
            self.subdiv_hpx.append(
                ahpx.HEALPix(nside=nside, order="nested", frame=acrd.ICRS())
            )

        self.indexed = False
        if load_index:
            self._load_index()
        else:
            self.pixels = self._traverse()
        self._pixcache = None

    def _traverse(self):
        n_levels = len(self.subdivs) + 1

        # Traverse the skymap tree, find all full-qualified healpix
        fqhipix = glob.glob(os.path.join(self.root, "/".join(["??????"] * n_levels)))

        pixels = {}
        for fqi in fqhipix:
            ipix = str(int(fqi.split("/")[-1]))
            pixels[ipix] = {"path": fqi}
        return pixels

    def _index_ipix(self, pix):
        nights = glob.glob(os.path.join(self.pixels[pix]["path"], "????????.fits"))
        pixel_dict = {}
        pixel_dict[pix] = self.pixels[pix]
        pixel_dict[pix]["nights"] = {}
        for night in nights:
            if self._pixcache is not None and night in self._pixcache[pix]["nights"]:
                pixel_dict[pix]["nights"][night] = self._pixcache[pix]["nights"][night]
                continue
            night_dict = {
                "ntiles": 0,
                "ntiles_zweighted": 0,
                "filenames": [],
                "epochs": [],
                "data_start_byte": [],
                "data_size": [],
            }

            ahdulist = arf.open(night, get_byte_positions=True)
            tile_size = ahdulist[1].data.size
            for tile in ahdulist[1:]:
                night_dict["ntiles"] += 1
                night_dict["ntiles_zweighted"] += (
                    np.count_nonzero(tile.data) / tile_size
                )
                night_dict["filenames"].append(tile.header["ORIGNAME"])
                night_dict["epochs"].append(tile.header["DATE-OBS"])
                night_dict["data_start_byte"].append(tile.header["_BYTE_START"])
                night_dict["data_size"].append(tile.header["_NBYTES"])

            pixel_dict[pix]["nights"][night] = night_dict
        return pixel_dict

    def index(self):
        manifest_path = os.path.join(self.root, "manifest.json")
        summary_path = os.path.join(self.root, "summary.parquet")

        if os.path.exists(manifest_path):
            with open(manifest_path, "r") as f:
                self._pixcache = orjson.loads(f.read())
        else:
            self._pixcache = None

        pix_ids = sorted(list(self.pixels.keys()))
        n_ipix = len(pix_ids)

        pool = mp.Pool(config.INDEX_WORKERS)

        pixel_dicts = []
        with click.progressbar(
            pool.imap_unordered(self._index_ipix, pix_ids),
            label="Indexing: ",
            length=len(pix_ids),
        ) as pbar:
            for p in pbar:
                pixel_dicts.append(p)

        self.pixels = reduce(operator.ior, pixel_dicts, {})
        self._pixcache = None

        pool.close()
        pool.join()

        for pix in self.pixels:
            total_tiles = 0
            total_tiles_weighted = 0
            for night in self.pixels[pix]["nights"]:
                total_tiles += self.pixels[pix]["nights"][night]["ntiles"]
                total_tiles_weighted += self.pixels[pix]["nights"][night][
                    "ntiles_zweighted"
                ]
            self.pixels[pix]["total_tiles"] = total_tiles
            self.pixels[pix]["total_tiles_weighted"] = total_tiles_weighted

        with open(manifest_path, "w") as f:
            f.write(orjson.dumps(self.pixels).decode("utf-8"))

        # Finally, write a summary table
        t = np.zeros(n_ipix, dtype=int)
        wt = np.zeros(n_ipix, dtype=np.float32)
        ipix = np.zeros(n_ipix, dtype=int)
        ra = np.zeros(n_ipix, dtype=np.float32)
        dec = np.zeros(n_ipix, dtype=np.float32)

        fast_healpixer = fast_healpix.FastHealpix()

        for i, pix in enumerate(self.pixels):
            t[i] = self.pixels[pix]["total_tiles"]
            wt[i] = self.pixels[pix]["total_tiles_weighted"]
            ipix[i] = pix
            ra[i], dec[i] = fast_healpixer.healpix_to_radec(
                int(pix), self.nside_base, 0.0, 0.0
            )

        summary_table = pa.Table.from_pydict(
            {
                "ipix": ipix,
                "ra": ra,
                "dec": dec,
                "n_tiles": t,
                "n_tiles_density_weighted": wt,
            }
        )
        pq.write_table(summary_table, summary_path)
        self.indexed = True

        return self.pixels

    def _load_index(self):
        manifest_path = os.path.join(self.root, "manifest.json")
        if not os.path.isfile(manifest_path):
            warnings.warn("No manifest file found! Do you need to index this archive?")
            return False
        with open(os.path.join(self.root, "manifest.json"), "rb") as f:
            manifest = orjson.loads(f.read())
        self.indexed = True
        self.pixels = manifest

    def get_ipix_path(self, ipix):
        fast_healpixer = fast_healpix.FastHealpix()

        ra, dec = fast_healpixer.healpix_to_radec(
            ipix,
            self.nside_base,
            0.0,
            0.0,
        )

        path = self.root
        for nside in self.subdivs:
            subhealpix = fast_healpixer.radec_to_healpix(ra, dec, nside)

            path = os.path.join(path, f"{subhealpix:06}")
        path = os.path.join(path, f"{ipix:06}")
        return path

    def index_to_dataframe(self):
        flat_dict = {
            "ipix": [],
            "filename": [],
            "startbyte": [],
            "bytesize": [],
            "epochs": [],
        }
        for ipix in self.pixels:
            nights = self.pixels[ipix]["nights"]
            for night in nights:
                epochs = self.pixels[ipix]["nights"][night]["epochs"]
                for i in range(len(epochs)):
                    flat_dict["ipix"].append(ipix)
                    flat_dict["filename"].append(night)
                    flat_dict["startbyte"].append(
                        self.pixels[ipix]["nights"][night]["data_start_byte"][i]
                    )
                    flat_dict["bytesize"].append(
                        self.pixels[ipix]["nights"][night]["data_size"][i]
                    )
                    flat_dict["epochs"].append(
                        self.pixels[ipix]["nights"][night]["epochs"][i]
                    )

        df = pa.Table.from_pydict(flat_dict).to_pandas()
        return df



__all__ = ["EFTEAlertReceiver", "EFTEAlertStreamer"]

import base64
import io
import os
from typing import Any, Dict, List, Optional
from uuid import uuid4

import astropy.table as tbl
import blosc
import fastavro as fa
import orjson
import pandas as pd
from confluent_kafka import KafkaException

from .. import config, get_logger
from ..consumer import Consumer
from ..producer import Producer

PATH = os.path.realpath(os.path.dirname(__file__))


class EFTEAlertStreamer(Producer):
    def __init__(self) -> None:
        """
        Initialize the EFTEAlertStreamer instance.

        It inherits from the Producer class and configures the Kafka connection parameters
        using values from the config dictionary.

        This class is used for streaming raw candidate detections from the
        observatory to the Rico Kafka cluster.

        The host, port, and topic are retrieved from the config dictionary
        using the keys 'KAFKA_ADDR', 'KAFKA_PORT', and 'HBEAT_TOPIC' respectively.
        """
        super().__init__(
            host=config.KAFKA_ADDR,
            port=config.KAFKA_PORT,
        )
        self.topic_base = config.EFTE_TOPIC_BASE

        self.parsed_alert_schema = fa.schema.load_schema(
            f"{PATH}/schemas/efte.alert.avsc"
        )

    def _parse_catalog(
        self, catalog: tbl.Table, xmatches: List[Dict[str, List]]
    ) -> bytes:
        """
        Parses a catalog file and returns the serialized Avro data.

        Args:
            catalog_path (str): The path to the catalog file.

        Returns:
            bytes: The serialized Avro data.
            dict: Catalog metadata.
        """
        tab = catalog

        if "MJD" not in tab.meta:
            tab.meta["MJD"] = 60000.1
        if "CCDDETID" not in tab.meta:
            tab.meta["CCDDETID"] = "ML3103817"

        mjd = tab.meta["MJD"]
        camera_id = tab.meta["CCDDETID"]

        records = []
        for i, r in enumerate(tab):
            alert_data = {
                "schemavsn": self.parsed_alert_schema["version"],
                "publisher": "rico.efte_generator",
                "objectId": str(uuid4()),
            }

            stamp = blosc.compress(r["stamp"].data.tobytes())

            candidate = dict(r)
            candidate["stamp_bytes"] = stamp
            candidate["epoch"] = mjd
            candidate["camera"] = camera_id

            alert_data["candidate"] = candidate
            alert_data["xmatch"] = xmatches[i]

            records.append(alert_data)

        fo = io.BytesIO()
        fa.writer(fo, self.parsed_alert_schema, records)
        fo.seek(0)

        return fo.read()

    def push_alert(self, catalog: tbl.Table, xmatches: List[Dict[str, List]]) -> None:
        """
        Pushes data from a catalog file to a camera-specific topic.

        Args:
            catalog_path (str): The path to the catalog file.

        Returns:
            None

        """
        avro_data = self._parse_catalog(catalog, xmatches)
        topic = self.topic_base
        self.send_binary(avro_data, topic=topic)


class EFTEAlertReceiver(Consumer):
    def __init__(
        self, group: str, output_path: str, filter_path: Optional[str] = None
    ) -> None:
        """Initialize the EFTEAlertReceiver class.

        Args:
            group (str): The Kafka consumer group ID.
            output_path (str): The path where filtered candidate data will be written.
            filter_path (Optional[str]): The path to the text file containing filter conditions for xmatch data.
                If provided, candidates will be filtered based on the conditions in the file.
                Each condition should be in the format: 'column_name operator value'.
        """
        super().__init__(
            host=config.KAFKA_ADDR,
            port=config.KAFKA_PORT,
            topic=config.EFTE_TOPIC_BASE,
            group_id=group,
        )

        self.parsed_alert_schema = fa.schema.load_schema(
            f"{PATH}/schemas/efte.alert.avsc"
        )
        self.filter_path = filter_path
        self.output_path = output_path
        self.log = get_logger(__name__)

    def poll_and_record(self) -> None:
        """Start polling for Kafka messages and process the candidates based on filter conditions."""
        c = self.get_consumer()

        c.subscribe(
            [
                self.topic,
            ]
        )
        try:
            while True:
                event = c.poll(1.0)
                if event is None:
                    continue
                if event.error():
                    raise KafkaException(event.error())
                else:
                    alerts = self._decode(event.value())
                    self._filter_to_disk(alerts)

        except KeyboardInterrupt:
            print("Canceled by user.")
        finally:
            c.close()

    def _decode(self, message: bytes) -> List[Dict[str, Any]]:
        """Decode the AVRO message into a list of dictionaries.

        Args:
            message (bytes): The AVRO message received from Kafka.

        Returns:
            List[Dict[str, Any]]: A list of candidate dictionaries.
        """
        stringio = io.BytesIO(message)
        stringio.seek(0)

        records = []
        for record in fa.reader(stringio):
            records.append(record)
        return records

    def _write_candidate(self, alert: Dict[str, Any]) -> None:
        """Write the candidate data to a JSON file.

        Args:
            candidate (Dict[str, Any]): The candidate dictionary to be written to the JSON file.
        """
        alert["candidate"]["stamp_bytes"] = base64.b64encode(
            alert["candidate"]["stamp_bytes"]
        ).decode("utf-8")
        with open(
            os.path.join(self.output_path, f"{alert['objectId']}.json"), "wb"
        ) as f:
            f.write(orjson.dumps(alert))
        self.log.info(f'New candidate: {alert["objectId"]}.json')

    def _filter_to_disk(self, alerts: List[Dict[str, Any]]) -> None:
        """Filter the candidates based on the specified filter conditions.

        Args:
            candidates (List[Dict[str, Any]]): The list of candidate dictionaries to be filtered.

        Note:
            If the 'filter_path' is provided, the candidates will be filtered based on the conditions
            specified in the text file. If 'xmatch' field is empty or 'filter_path' is not provided,
            all candidates will be written to the output.
        """
        if self.filter_path is not None:
            with open(self.filter_path, "r") as file:
                filter_conditions = file.read().splitlines()
            filter_conditions = [f for f in filter_conditions if len(f) > 3]

        for alert in alerts:
            if self.filter_path is not None:
                if len(alert["xmatch"]) > 0:
                    xmatch = pd.DataFrame.from_records(alert["xmatch"])
                    xmatch["g-r"] = xmatch["g"] - xmatch["r"]
                    for condition in filter_conditions:
                        column_name, operator, value = condition.split()
                        xmatch = xmatch.query(f"{column_name} {operator} {value}")
                    if len(xmatch) > 0:
                        self._write_candidate(alert)
                else:
                    return
            else:
                self._write_candidate(alert)



__all__ = ["EFTEWatcher", "EFTECatalogHandler"]

import os
import time
from collections import defaultdict

import ray
from watchdog.events import FileSystemEvent, FileSystemEventHandler
from watchdog.observers.polling import PollingObserver

from .. import get_logger
from .processor import EFTECatalogProcessor


class EFTEWatcher:
    def __init__(self, watch_path: str) -> None:
        """Initialize the EFTEWatcher class.

        Args:
            watch_path (str): The path to the directory to watch for catalog files.
            format (str, optional): The format of catalog files. Defaults to "fits".
        """
        self.watch_path = os.path.abspath(watch_path)

    def watch(self) -> None:
        """Start watching the specified directory for catalog files."""
        ray.init()
        event_handler = EFTECatalogHandler()
        observer = PollingObserver()

        observer.schedule(event_handler, self.watch_path, recursive=False)
        observer.start()

        try:
            while True:
                time.sleep(1)
        except KeyboardInterrupt:
            observer.stop()
        observer.join()


class EFTECatalogHandler(FileSystemEventHandler):
    def __init__(self):
        """Initialize the EFTECatalogHandler class."""
        self.efte_processors = defaultdict(EFTECatalogProcessor.remote)
        self.log = get_logger(__name__)

    def on_created(self, event: FileSystemEvent) -> None:
        """Process the newly created catalog file.

        Args:
            event (FileSystemEvent): The event object representing the file creation.

        Returns:
            None: This method does not return any value; it processes the catalog file.
        """
        filepath = event.src_path

        if filepath[-4:] != ".cat":
            return
        camera_id = os.path.basename(filepath)[:9]

        self.log.info(f"New cat for {camera_id}: {filepath}")

        self.efte_processors[camera_id].process.remote(filepath)



__all__ = ["MissingDirectoryError", "ATLASRefcat2"]

import glob
import os
from concurrent.futures import ThreadPoolExecutor
from typing import List, Tuple, Union

import astropy.coordinates as crds
import astropy.units as u
import click
import numpy as np
import pandas as pd
import pyarrow as pa
import pyarrow.dataset as pds
import pyarrow.feather as pf
import pyarrow.parquet as pq
from astropy_healpix import HEALPix

from .. import config, s3


class MissingDirectoryError(Exception):
    """Exception raised when a directory is missing.

    Attributes:
        message (str): Explanation of the error.
    """

    def __init__(self, message: str = "Tree basepath is not set!"):
        self.message = message
        super().__init__(self.message)


def haversine(
    lon1: np.ndarray, lat1: np.ndarray, lon2: np.ndarray, lat2: np.ndarray
) -> np.ndarray:
    """Calculate the great circle distance between two points on the earth.

    Args:
        lon1 (np.ndarray): Longitudes of the first points in decimal degrees.
        lat1 (np.ndarray): Latitudes of the first points in decimal degrees.
        lon2 (np.ndarray): Longitudes of the second points in decimal degrees.
        lat2 (np.ndarray): Latitudes of the second points in decimal degrees.

    Returns:
        np.ndarray: The great circle distances in degrees.
    """
    lon1, lat1, lon2, lat2 = map(np.radians, [lon1, lat1, lon2, lat2])

    dlon = lon2 - lon1
    dlat = lat2 - lat1

    a = np.sin(dlat / 2.0) ** 2 + np.cos(lat1) * np.cos(lat2) * np.sin(dlon / 2.0) ** 2

    c = 2 * np.arcsin(np.sqrt(a))
    return np.rad2deg(c)


class ATLASRefcat2:
    def __init__(self):
        self.h4 = HEALPix(nside=4, order="nested", frame=crds.ICRS())
        self.h16 = HEALPix(nside=16, order="nested", frame=crds.ICRS())

        self.table: pa.Table = None
        self.dataset: str = None

        if os.path.isdir(os.path.join("/etc/rico/atlas_refcat2")):
            self.dataset = "/etc/rico/atlas_refcat2"

        else:
            if not os.path.isdir(os.path.join(config.RICO_CACHE_DIR, "atlas_refcat2")):
                if click.confirm(
                    "ATLAS RefCat not found, do you want to download it (82 GB)?"
                ):
                    s3share = s3.S3Share()
                    s3share.download_atlas()

                else:
                    raise MissingDirectoryError("ATLAS Refcat is not installed")
            self.dataset = os.path.join(config.RICO_CACHE_DIR, "atlas_refcat2")

    def radial(
        self,
        ra: float,
        dec: float,
        radius: float,
        min_g: float = 11.0,
        max_g: float = 16.0,
        return_area: bool = False,
        grab_closest: bool = False,
        as_pandas: bool = True,
    ) -> Union[pd.DataFrame, pa.Table]:
        """Perform radial search on the dataset.

        Args:
            ra (float): Right ascension in degrees.
            dec (float): Declination in degrees.
            radius (float): Radius of the search in degrees.
            min_g (float, optional): Minimum 'g' value for filtering. Defaults to 11.0.
            max_g (float, optional): Maximum 'g' value for filtering. Defaults to 16.0.
            return_area (bool, optional): Whether to return the area covered by the search.
                Defaults to True.
            grab_closest (bool, optional): Whether to return only the closest match.
                Defaults to False.
            as_pandas (bool, optional): Whether to return the result as a pandas DataFrame.
                If False, the result is returned as a pyarrow Table. Defaults to True.

        Returns:
            Union[pd.DataFrame, pa.Table]: The filtered dataset within the specified radius.
        """
        hpx_4 = self.h4.cone_search_lonlat(ra * u.deg, dec * u.deg, radius * u.deg)
        hpx_16 = self.h16.cone_search_lonlat(ra * u.deg, dec * u.deg, radius * u.deg)

        imfiles = []
        for i4 in hpx_4:
            files_4 = [
                glob.glob(
                    os.path.join(
                        self.dataset, str(i4) + "/" + str(i) + "/" + "*.feather"
                    )
                )
                for i in hpx_16
            ]
            imfiles += files_4
        imfiles = [x[0] for x in imfiles if len(x) > 0]

        if as_pandas:
            with ThreadPoolExecutor() as threads:
                t_res = threads.map(
                    self._from_featherfile_pandas,
                    zip(
                        imfiles,
                        [
                            [min_g, max_g],
                        ]
                        * len(imfiles),
                    ),
                )
            t = pd.concat(t_res)
            separations = haversine(t["ra"], t["dec"], ra, dec)
            t["separation"] = separations

            if grab_closest:
                t = t.iloc[[np.argmin(separations)]]
            else:
                t = t[separations < radius]

        else:
            with ThreadPoolExecutor() as threads:
                t_res = threads.map(
                    self._from_featherfile,
                    zip(
                        imfiles,
                        [
                            [min_g, max_g],
                        ]
                        * len(imfiles),
                    ),
                )

            t = pa.concat_tables(t_res)
            separations = haversine(t["ra"].to_numpy(), t["dec"].to_numpy(), ra, dec)
            t.append_column("separation", [pa.array(separations)])

            t = t.filter(separations < radius)

        if return_area:
            return t, np.pi * radius**2
        else:
            return t

    @staticmethod
    def _from_featherfile_pandas(packet: Tuple[str, List[float]]) -> pd.DataFrame:
        """Read a feather file and return the DataFrame after filtering.

        Args:
            packet (Tuple[str, List[float]]): A tuple containing the feather file path
                and the range of 'g' values for filtering.

        Returns:
            pd.DataFrame: The filtered DataFrame.
        """
        path, [min_g, max_g] = packet
        t = pf.read_feather(path)

        t = t[(t["g"] < max_g) & (t["g"] > min_g)]

        return t

    @staticmethod
    def _from_featherfile(packet: Tuple[str, List[float]]) -> pa.Table:
        """Read a feather file and return the Table.

        Args:
            packet (Tuple[str, List[float]]): A tuple containing the feather file path
                and dummy list.

        Returns:
            pa.Table: The Table read from the feather file.
        """
        path, [_, _] = packet
        t = pf.read_table(path)

        return t

    def to_parquet(self, outpath: str) -> None:
        """Write the Table to a parquet file.

        Args:
            outpath (str): The output path for the parquet file.
        """
        pq.write_table(self.table, outpath)

    def from_parquet(self, path: str) -> None:
        """Load a parquet file into the class.

        Args:
            path (str): The path to the parquet file.

        Raises:
            FileNotFoundError: If the parquet file at the given path is not found.
        """
        try:
            self.table = pq.read_table(path)
        except FileNotFoundError:
            raise FileNotFoundError(f"Parquet file not found at path: {path}")

    def to_ds(self, outdir: str) -> None:
        """Write the Table to a dataset in feather format.

        Args:
            outdir (str): The output directory for the dataset.
        """
        part = pds.partitioning(
            pa.schema(
                [
                    ("h4", pa.int32()),
                    ("h16", pa.int32()),
                    # ("h32", pa.int32()),
                    # ("h64", pa.int32()),
                    # ("h256", pa.int32()),
                ]
            ),
        )
        pds.write_dataset(
            self.table,
            outdir,
            format="feather",
            max_partitions=786432,
            partitioning=part,
            max_open_files=786432,
        )

    def to_segment_ds(self, outdir: str, nside_base: int) -> None:
        """Write the Table to a segmented dataset in feather format.

        Args:
            outdir (str): The output directory for the segmented dataset.
            nside_base (int): The base nside value for segmentation.
        """
        part = pds.partitioning(
            pa.schema(
                [
                    ("h4", pa.int32()),
                    ("h16", pa.int32()),
                    ("h64", pa.int32()),
                    ("h256", pa.int32()),
                ]
            ),
        )
        pds.write_dataset(
            self.table,
            outdir,
            format="feather",
            max_partitions=786432,
            partitioning=part,
            max_open_files=786432,
        )



__all__ = ["VetNet"]
"""Wrapper for the MC EFTE Vetnet model."""
import os
from typing import Tuple

os.environ["TF_CPP_MIN_LOG_LEVEL"] = "2"

import astropy.visualization as viz  # noqa: E402
import numpy as np  # noqa: E402
import tensorflow as tf  # noqa: E402
import tensorflow.keras.models as models  # noqa: E402

from .. import s3  # noqa: E402


class VetNet:
    def __init__(self) -> None:
        """Initialize the Vetnet class.

        Loads the pretrained model and sets class variables.
        """
        try:
            if not os.path.isdir(os.path.join(os.path.expanduser("~"), ".rico_cache")):
                os.mkdir(os.path.join(os.path.expanduser("~"), ".rico_cache"))

            self.model = models.load_model(
                os.path.join(os.path.expanduser("~"), ".rico_cache", "hypersky_v7_v0")
            )
        except OSError:
            s3share = s3.S3Share()
            s3share.download_vetnet()
            self.model = models.load_model(
                os.path.join(os.path.expanduser("~"), ".rico_cache", "hypersky_v7_v0")
            )

        self.BATCH_SIZE: int  # The batch size for data processing
        self.AUTOTUNE: int = tf.data.AUTOTUNE  # The buffer size for prefetching
        self.zed = viz.ZScaleInterval()

    def prep_ds(self, ds: tf.data.Dataset) -> tf.data.Dataset:
        """Preprocess the input dataset.

        Batch the dataset and apply buffered prefetching.

        Args:
            ds (tf.data.Dataset): The input dataset.

        Returns:
            tf.data.Dataset: The preprocessed dataset.
        """
        # Batch all datasets
        ds = ds.batch(self.BATCH_SIZE)

        # Use buffered prefetching on all datasets
        return ds.prefetch(buffer_size=self.AUTOTUNE)

    def _normalize_triplet(self, data: np.ndarray) -> np.ndarray:
        """Perform stamp normalization. Processes the reference, science, and
        difference images separately.


        Args:
            data (np.ndarray): The input data for prediction.

        Returns:
            np.ndarray: Recursively normalized postage stamp cutout.
        """
        for i in range(data.shape[-1]):
            data[:, :, i] = self.zed(data[:, :, i])

        return data

    def _normalize_stamps(self, data: np.ndarray) -> np.ndarray:
        """Perform recursive stamp normalization. Processes the reference,
        science, and difference images separately.


        Args:
            data (np.ndarray): The input data for prediction.

        Returns:
            np.ndarray: Recursively normalized postage stamp cutouts.
        """
        for i in range(data.shape[0]):
            data[i, :, :, :] = self._normalize_triplet(data[i, :, :, :])

        return data

    def mc_predict(
        self, data: np.ndarray, n_posterior_draws: int
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        """Perform Monte Carlo prediction.

        Args:
            data (np.ndarray): The input data for prediction.
            n_posterior_draws (int): The number of posterior draws.

        Returns:
            Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]: A tuple containing:
            - mean_pred (np.ndarray): The mean prediction.
            - low_pred (np.ndarray): The lower prediction bound.
            - high_pred (np.ndarray): The upper prediction bound.
            - confidence (np.ndarray): The confidence measure.
        """
        data = self._normalize_stamps(data)

        pred_dists = np.array(
            [self.model(data, training=False) for _ in range(n_posterior_draws)]
        )

        pred_dists[pred_dists == 0] += 1e-5
        pred_dists[pred_dists == 1] -= 1e-5

        entropy = (-1 * pred_dists * np.log2(pred_dists)) - (
            (1 - pred_dists) * np.log2(1 - pred_dists)
        )
        low_pred = np.rint(np.percentile(pred_dists, (50.0 - 95.4 / 2), axis=0))
        high_pred = np.rint(np.percentile(pred_dists, (50.0 + 95.4 / 2), axis=0))
        mean_pred = np.rint(np.mean(pred_dists, axis=0))

        confidence = 1 - (1 / n_posterior_draws) * np.sum(entropy, axis=0)

        return mean_pred, low_pred, high_pred, confidence



__all__ = ["insert_efte_candidates"]

import datetime
import os

import astropy.table as tbl
import astropy.time as atime
import astropy.visualization as viz
import numpy as np
import psycopg
import skimage as sk

from . import config, get_logger

log = get_logger("__name__")


def insert_efte_candidates(catalog):
    if len(catalog) == 0:
        return False

    conn = psycopg.connect(
        f"""host='{config.EFTE_DB_ADDR}'
            port='{config.EFTE_DB_PORT}'
            dbname='{config.EFTE_DB_NAME}'
            user='{config.EFTE_DB_USER}'
            password='{config.EFTE_DB_PASS}'
        """,
        # cursor_factory=psycopg.ClientCursor,
    )

    c = conn.cursor()

    now = datetime.datetime.utcnow()
    imgepoch = atime.Time(catalog.meta["EPOCH"], format="isot", scale="utc")
    epoch = [imgepoch.to_datetime()] * len(catalog)
    lag = [np.abs((now - imgepoch.to_datetime()).total_seconds())] * len(catalog)
    camera = [catalog.meta["CCD"]] * len(catalog)
    ratchet = [catalog.meta["ratchetnum"]] * len(catalog)
    filename = [catalog.meta["FILENAME"]] * len(catalog)

    zed = viz.ZScaleInterval()
    stamps = [
        sk.img_as_ubyte(zed(x)).astype(np.uint8).tobytes()
        for x in catalog["stamp"].data
    ]
    if "bstamps" not in catalog.colnames:
        catalog.add_column(tbl.Column(stamps, name="bstamps"))

    ra = catalog["ra"].data.astype(float)
    dec = catalog["dec"].data.astype(float)
    x = catalog["xcentroid"].data.astype(float)
    y = catalog["ycentroid"].data.astype(float)
    sign_ratio = catalog["sign_ratio"].data.astype(float)
    srcsnr = catalog["srcsnr"].data.astype(float)
    gmag = catalog["gmag"].data.astype(float)
    difsnr = catalog["difsnr"].data.astype(float)
    vetscore = catalog["vetnet"].data.astype(float)
    stamps = catalog["bstamps"].data
    site = ["mlo"] * len(stamps)

    rows = tuple(zip(epoch, stamps))

    try:
        q = """
               INSERT INTO stamps (epoch, cutout)
               VALUES (%s, %s)
               RETURNING id"""
        c.executemany(q, rows, returning=True)
        stamp_ids = []
        while True:
            stamp_ids.append(c.fetchone()[0])
            if not c.nextset():
                break

    except Exception as e:
        log.error(
            "Stamp insert failed for {}  aborting: {} {}".format(
                os.path.basename(catalog.meta["FILENAME"]), q, e
            )
        )

        c.close()
        conn.close()
        return False

    rows = list(
        zip(
            epoch,
            ra,
            dec,
            x,
            y,
            sign_ratio,
            vetscore,
            srcsnr,
            difsnr,
            gmag,
            camera,
            ratchet,
            filename,
            lag,
            site,
            stamp_ids,
        )
    )

    q = """
            INSERT INTO  candidates(
                               epoch,
                               raj2000,
                               decj2000,
                               x_position,
                               y_position,
                               sign_ratio,
                               vetnet_score,
                               srcsnr,
                               difsnr,
                               mag_g,
                               camera,
                               ratchet,
                               detection_image,
                               detection_lag,
                               detected_from,
                               stamp_id)
            VALUES (%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s)
            """
    c.executemany(q, rows)

    c.close()
    conn.commit()
    conn.close()



import sys

import cffi
import numpy as np

from .fast_healpix_c import lib as fast_healpix_mod  # noqa: E402


class FastHealpix:
    def __init__(self, recompile=False, max_call_len=10000):
        # compile the C code and import the module
        self.ffi = cffi.FFI()

        self.max_call_len = max_call_len

        self.ras = np.zeros(max_call_len, dtype=np.float64)
        self.decs = np.zeros(max_call_len, dtype=np.float64)
        self.healpixes = np.zeros(max_call_len, dtype=np.int64)
        self.ra_ret = self.ffi.cast("double *", self.ras.ctypes.data)
        self.dec_ret = self.ffi.cast("double *", self.decs.ctypes.data)
        self.healpixes_ret = self.ffi.cast("long int *", self.healpixes.ctypes.data)

    def healpixes_to_radec(self, healpixes, nside, dx, dy):
        npix = len(healpixes)
        if npix > self.max_call_len:
            print("Exceeded maximum list length in healpixes_to_radec")
            sys.exit(1)

        healpixes = np.array(healpixes, dtype=np.int64)
        healpixes_c = self.ffi.cast("long int *", healpixes.ctypes.data)

        fast_healpix_mod.healpixls_to_radecdeg(
            healpixes_c, npix, nside, dx, dy, self.ra_ret, self.dec_ret
        )

        return self.ras[:npix], self.decs[:npix]

    def healpix_deltagrid_to_radec(self, healpix, nside, dxs, dys):
        npoints = len(dxs)
        if npoints > self.max_call_len:
            print("Exceeded maximum list length in healpixes_to_radec")
            sys.exit(1)

        dxs = np.array(dxs, dtype=np.float64)
        dys = np.array(dys, dtype=np.float64)
        dxs_c = self.ffi.cast("double *", dxs.ctypes.data)
        dys_c = self.ffi.cast("double *", dys.ctypes.data)

        fast_healpix_mod.healpixl_grid_to_radecdeg(
            healpix, npoints, nside, dxs_c, dys_c, self.ra_ret, self.dec_ret
        )

        return self.ras[:npoints], self.decs[:npoints]

    def healpix_to_radec(self, healpix, nside, dx, dy):
        fast_healpix_mod.healpixl_to_radecdeg(
            healpix, nside, dx, dy, self.ra_ret, self.dec_ret
        )

        return self.ras[0], self.decs[0]

    def radec_to_healpix(self, ra, dec, nside):
        return fast_healpix_mod.radec_to_healpixl(ra, dec, nside)

    def radecs_to_healpixes(self, ras, decs, nside):
        nras = len(ras)
        if nras > self.max_call_len:
            print("Exceeded maximum list length in radecs_to_healpixes")
            sys.exit(1)

        ras = np.array(ras, dtype=np.float64)
        decs = np.array(decs, dtype=np.float64)
        ras_c = self.ffi.cast("double *", ras.ctypes.data)
        decs_c = self.ffi.cast("double *", decs.ctypes.data)

        fast_healpix_mod.radecs_to_healpixls(
            ras_c, decs_c, nras, self.healpixes_ret, nside
        )

        return self.healpixes[:nras]



__all__ = ["Timer"]

import time


class Timer:
    def __init__(self, log=None):
        self.start_time = time.perf_counter()
        self.last_ping = time.perf_counter()
        self.log = log

    def start(self):
        """Total runtime start."""
        self.start_time = time.perf_counter()

    @property
    def t(self):
        """Runtime relative to start."""
        return time.perf_counter() - self.start_time

    @property
    def t_since_last_ping(self):
        """Runtime relative to the last ping."""
        return time.perf_counter() - self.last_ping

    def ping(self, message):
        if self.log is not None:
            self.log.info(
                f"@ {self.t * 1000:0.0f} ms ({self.t_since_last_ping * 1000:0.0f} ms delta): {message}",
            )
        self.last_ping = time.perf_counter()



"""Ray-parallelizable EFTE catalog reducer."""
import os

import astropy.table as tbl
import ray

from .. import catalogs, get_logger, utils
from .stream import EFTEAlertStreamer
from .vetnet import VetNet


@ray.remote
class EFTECatalogProcessor:
    def __init__(self):
        """Initialize the EFTECatalogProcessor class."""
        self.vetnet = VetNet()
        self.atlas = catalogs.ATLASRefcat2()
        self.producer = EFTEAlertStreamer()

        self.log = get_logger(__name__)

    def process(self, filepath: str) -> None:
        """Perform vetting and crossmatching for the given FITS table.

        Args:
            filepath (str): The path to the FITS table.

        Returns:
            Table: Returns the FITS table with normalized stamps, scores.
        """
        clock = utils.Timer(log=self.log)
        name = os.path.basename(filepath)
        table: tbl.Table = tbl.Table.read(filepath, format="fits")
        clock.ping(f"Read {len(table)} candidates from {name}")

        stamps = table["stamp"].data
        mean_pred, _, _, confidence = self.vetnet.mc_predict(stamps, 10)
        clock.ping(f"Vetted candidates from {name}")

        table["vetnet_score"] = confidence[:, 0]
        table = table[mean_pred[:, 0] > 0.5]
        table = table[table["vetnet_score"] > 0.4]  # fairly arbitrary...

        clock.ping(f"Reporting {len(table)} candidates in {name}")

        if len(table) == 0:
            return
        crossmatches = []
        for r in table:
            phot = self.atlas.radial(
                r["ra"], r["dec"], 0.01, max_g=25, return_area=False
            )
            phot = phot.sort_values(by="separation")
            crossmatches.append(phot.to_dict(orient="records"))

        clock.ping(f"Xmatched {name} with ATLAS Refcat")

        self.producer.push_alert(table, crossmatches)



import base64
from typing import List, Type, TypeVar

import blosc
import numpy as np
import orjson
from pydantic import BaseModel

EFTEAlertType = TypeVar("EFTEAlertType", bound="EFTEAlert")


class EFTECandidate(BaseModel):
    stamp_bytes: bytes
    epoch: float
    camera: str
    thresh: float
    vetnet_score: float
    npix: int
    tnpix: int
    xmin: int
    xmax: int
    ymin: int
    ymax: int
    xcentroid: float
    ycentroid: float
    x2: float
    y2: float
    xy: float
    errx2: float
    erry2: float
    errxy: float
    a: float
    b: float
    theta: float
    cxx: float
    cyy: float
    cxy: float
    cflux: float
    flux: float
    cpeak: float
    peak: float
    xcpeak: int
    ycpeak: int
    xpeak: int
    ypeak: int
    flag: int
    aper_sum_bkgsub: float
    annulus_rms: float
    difsnr: float
    hfd: float
    sign_ratio: float
    insmag: float
    zpoint: float
    gmag: float
    srcsnr: float
    ra: float
    dec: float

    @property
    def stamp(self):
        return (
            np.frombuffer(
                blosc.decompress(
                    base64.b64decode(
                        self.stamp_bytes,
                    )
                ),
                dtype=np.float32,
            )
            .reshape((30, 30, 3))
            .byteswap()
        )


# Create a Pydantic model for the 'xmatch' list of dictionaries
class XMatchItem(BaseModel):
    ra: float
    dec: float
    parallax: float
    pmra: float
    pmdec: float
    g: float
    r: float
    i: float
    separation: float


class EFTEAlert(BaseModel):
    schemavsn: str
    publisher: str
    objectId: str
    candidate: EFTECandidate
    xmatch: List[XMatchItem]

    @classmethod
    def from_json(cls: Type[EFTEAlertType], json_path: str) -> EFTEAlertType:
        with open(json_path, "rb") as f:
            alert = orjson.loads(f.read())
        return cls(**alert)



from typing import Union

import confluent_kafka as ck

from . import get_logger


class Consumer:
    def __init__(
        self,
        host: str,
        port: Union[int, str],
        group_id: str,
        topic: str,
    ) -> None:
        """Initialize the Consumer class.

        Args:
            host (str): The Kafka broker host.
            port (Union[int, str]): The Kafka broker port.
            group_id (str): The Kafka consumer group ID.
            topic (str): The Kafka topic to consume messages from.
        """
        # Unlike producer, consumer is created on poll
        self.config = {
            "bootstrap.servers": f"{host}:{port}",
            "group.id": group_id,
            "enable.auto.commit": False,
            "auto.offset.reset": "latest",
        }

        self.log = get_logger(__name__)
        self.topic = topic
        self.polling = False

    def get_consumer(self) -> ck.Consumer:
        """Create and return a new Kafka consumer instance.

        Returns:
            ck.Consumer: A new Kafka consumer instance configured with the specified settings.
        """
        return ck.Consumer(self.config)
